@@ -2,10 +2,12 @@ import argparse
 import copy
 import collections
 import hashlib
+import importlib
 import json
 import logging
 import os
 import re
+import six
 import subprocess
 import sys
 import tempfile
@@ -15,7 +17,7 @@ from contextlib import contextmanager
 from .diff_match_patch import diff_match_patch
 import blessings
 import pathspec
-from path import path
+from path import Path as path
 
 log = logging.getLogger('utils')
 
@@ -45,17 +47,21 @@ def tempdir(chdir=True):
 
 def deepmerge(dest, src):
     """
-    Deep merge of two dicts.
+    Deep merge all the keys in src. When the value of the key is a dict,
+    recursively call deepmerge. When the value of the key is a list, iterate
+    the entire list and de-duplicate the list. When neither a dict or a list
+    call the python copy.deepcopy method.
 
-    This is destructive (`dest` is modified), but values
-    from `src` are passed through `copy.deepcopy`.
+    This is destructive (`dest` is modified), as values from `src` may be
+    passed through `copy.deepcopy`.
     """
-    for k, v in src.iteritems():
+    for k, v in src.items():
         if dest.get(k) and isinstance(v, dict):
             deepmerge(dest[k], v)
         elif dest.get(k) and isinstance(v, list):
-            if not v in dest.get(k):
-                dest[k].extend(v)
+            for item in v:
+                if item not in dest[k]:
+                    dest[k].append(item)
         else:
             dest[k] = copy.deepcopy(v)
     return dest
@@ -133,9 +139,15 @@ class ProcessResult(object):
     def output(self):
         result = ''
         if self.stdout:
-            result += self.stdout
+            stdout = self.stdout
+            if six.PY3:
+                stdout = stdout.decode('UTF-8')
+            result += stdout
         if self.stderr:
-            result += self.stderr
+            stderr = self.stderr
+            if six.PY3:
+                stderr = stderr.decode('UTF-8')
+            result += stderr
         return result.strip()
 
     @property
@@ -202,6 +214,7 @@ class Process(object):
             result.exit_on_error()
         return result
 
+
 command = Process
 
 
@@ -277,7 +290,7 @@ def retry(attempts, *callbacks, **kwargs):
     """
     Repeatedly try callbacks a fixed number of times or until all return True
     """
-    for attempt in xrange(attempts):
+    for attempt in range(attempts):
         if 'bar' in kwargs:
             kwargs['bar'].next(attempt == 0, message=kwargs.get('message'))
         for callback in callbacks:
@@ -307,25 +320,20 @@ def which(program):
     return None
 
 
-def load_class(dpath, workingdir=None):
-    # we expect the last element of the path
+def load_class(full_class_string, workingdir=None):
+    """
+    dynamically load a class from a string
+    """
+
+    module_path, class_str = full_class_string.rsplit(".", 1)
     if not workingdir:
         workingdir = os.getcwd()
-    with cd(workingdir):
-        modpath, classname = dpath.rsplit('.', 1)
-        modpath = path(modpath.replace(".", "/"))
-        if not modpath.exists():
-            modpath += ".py"
-        if not modpath.exists():
-            raise OSError("Unable to load {} from {}".format(
-                dpath, workingdir))
-        namespace = {}
-        execfile(modpath, globals(), namespace)
-        klass = namespace.get(classname)
-        if klass is None:
-            raise ImportError("Unable to load class {} at {}".format(
-                classname, dpath))
-        return klass
+    sys.path.insert(0, workingdir)
+    try:
+        module = importlib.import_module(module_path)
+        return getattr(module, class_str)
+    finally:
+        sys.path.pop(0)
 
 
 def walk(pathobj, fn, matcher=None, kind=None, **kwargs):
@@ -363,13 +371,13 @@ def sign(pathobj):
     return hashlib.sha256(p.bytes()).hexdigest()
 
 
-def delta_signatures(manifest_filename, ignorer=None):
+def delta_signatures(manifest_filename, ignore=None):
     md = path(manifest_filename)
     repo = md.normpath().dirname()
 
     expected = json.load(md.open())
     current = {}
-    for rel, sig in walk(repo, sign):
+    for rel, sig in walk(repo, sign, ignore_matcher(ignore or [])):
         rel = rel.relpath(repo)
         current[rel] = sig
     add, change, delete = set(), set(), set()
@@ -378,8 +386,6 @@ def delta_signatures(manifest_filename, ignorer=None):
         fp = repo / p
         if not fp.isfile():
             continue
-        if ignorer and not ignorer(p):
-            continue
 
         if p not in expected["signatures"]:
             add.add(p)
@@ -430,7 +436,7 @@ class TermWriter(object):
     def __getattr__(self, key):
         return getattr(self.term, key)
 
-    def write(self, msg,  *args, **kwargs):
+    def write(self, msg, *args, **kwargs):
         if 't' in kwargs:
             raise ValueError("Using reserved token 't' in TermWriter.write")
         kwargs['t'] = self.term
@@ -441,26 +447,35 @@ class _O(dict):
     def __getattr__(self, k):
         return self[k]
 
+
 REACTIVE_PATTERNS = [
-    re.compile("\s*@when"),
-    re.compile(".set_state\(")
+    re.compile(r"\s*@when"),
+    re.compile(r".set_state\(")
 ]
 
 
 def delta_python(orig, dest, patterns=REACTIVE_PATTERNS, context=2):
     """Delta two python files looking for certain patterns"""
-    if isinstance(orig, path):
-        od = orig.text()
-    elif hasattr(orig, 'read'):
-        od = orig.read()
-    else:
-        raise TypeError("Expected path() or file(), got %s" % type(orig))
-    if isinstance(dest, path):
-        dd = dest.text()
-    elif hasattr(orig, 'read'):
-        dd = dest.read()
-    else:
-        raise TypeError("Expected path() or file(), got %s" % type(dest))
+    try:
+        if isinstance(orig, path):
+            od = orig.text()
+        elif hasattr(orig, 'read'):
+            od = orig.read()
+        else:
+            raise TypeError("Expected path() or file(), got %s" % type(orig))
+    except Exception:
+        log.error('Error reading {}'.format(orig))
+        raise
+    try:
+        if isinstance(dest, path):
+            dd = dest.text()
+        elif hasattr(orig, 'read'):
+            dd = dest.read()
+        else:
+            raise TypeError("Expected path() or file(), got %s" % type(dest))
+    except Exception:
+        log.error('Error reading {}'.format(dest))
+        raise
 
     differ = diff_match_patch()
     linect = 0
@@ -529,7 +544,7 @@ def delta_python_dump(orig, dest, patterns=REACTIVE_PATTERNS,
 
 
 class Description(argparse._StoreTrueAction):
-    """A argparse action that prints its parent parser's description and exits."""
+    """A argparse action that prints parent parser's description & exits."""
     def __call__(self, parser, namespace, values, option_string=None):
         print(parser.description.split('\n')[0].strip('. '))
         raise SystemExit()
@@ -537,3 +552,84 @@ class Description(argparse._StoreTrueAction):
 
 def add_plugin_description(parser):
     parser.add_argument('--description', action=Description)
+
+
+class OrderedSet(collections.MutableSet):
+    def __init__(self, iterable=None):
+        self.end = end = []
+        end += [None, end, end]
+        self.map = {}
+        if iterable is not None:
+            self |= iterable
+
+    def __len__(self):
+        return len(self.map)
+
+    def __contains__(self, key):
+        return key in self.map
+
+    def add(self, key):
+        if key not in self.map:
+            end = self.end
+            curr = end[1]
+            curr[2] = end[1] = self.map[key] = [key, curr, end]
+
+    def discard(self, key):
+        if key in self.map:
+            key, prev, next = self.map.pop(key)
+            prev[2] = next
+            next[1] = prev
+
+    def __iter__(self):
+        end = self.end
+        curr = end[2]
+        while curr is not end:
+            yield curr[0]
+            curr = curr[2]
+
+    def __reversed__(self):
+        end = self.end
+        curr = end[1]
+        while curr is not end:
+            yield curr[0]
+            curr = curr[1]
+
+    def pop(self, last=True):
+        if not self:
+            raise KeyError('set is empty')
+        key = self.end[1][0] if last else self.end[2][0]
+        self.discard(key)
+        return key
+
+    def __repr__(self):
+        if not self:
+            return '%s()' % (self.__class__.__name__,)
+        return '%s(%r)' % (self.__class__.__name__, list(self))
+
+    def __eq__(self, other):
+        if isinstance(other, OrderedSet):
+            return len(self) == len(other) and list(self) == list(other)
+        return set(self) == set(other)
+
+
+def validate_display_name(entity, linter):
+    """Validate the display name info in entity metadata.
+
+    :param entity: dict of entity metadata parsed from metadata.yaml
+    :param linter: :class:`CharmLinter` or :class: `BundleLinter` object to
+        which info/warning/error messages will be written
+    """
+    if 'display-name' not in entity:
+        linter.info('`display-name` not provided, '
+                    'add for custom naming in the UI')
+        return
+
+    # Keep display name in sync with the juju/names package.
+    # https://github.com/juju/names/blob/master/charm.go#L33
+    match = re.match(
+        r'^[A-Za-z0-9]+( [-A-Za-z0-9_]|[-A-Za-z0-9_])*$',
+        entity['display-name'])
+    if not match:
+        linter.err('display-name: not in valid format. '
+                   'Only letters, numbers, dashes, and hyphens are permitted.')
+        return
